/*
 *  Copyright (c) 2008 Stephen Tarzia
 */

/*
 * Remote commander input driver for Linux
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Should you need to contact me, the author, you can do so either by
 * e-mail - mail your message to <spt2002@columbia.edu>, or by paper mail:
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/serio.h>
#define SERIO_REMOTECMD        0x35

#define DRIVER_DESC	"Remote Commander keyboard driver"

MODULE_AUTHOR("Stephen Tarzia <spt2002@columbia.edu>");
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

#define RCMD_KEY	0x7f
#define RCMD_PRESS	0x80

/* There are two classes of keycodes that Remote commander sends, control
 * keys and regular keys */

#define NUM_CTRL_CHARS 32
static unsigned char ctrl_keycode[NUM_CTRL_CHARS] = { 
  KEY_LEFTSHIFT, KEY_LEFTALT, KEY_LEFTCTRL, KEY_F1, 
  KEY_F2, KEY_F3, KEY_F4, KEY_F5, 
  KEY_F6, KEY_F7, KEY_F8, KEY_F9, 
  KEY_F10, KEY_F11, KEY_F12, 0,
  KEY_INSERT, KEY_DELETE, KEY_HOME, KEY_PAGEUP,
  KEY_PAGEDOWN, KEY_END, KEY_UP, KEY_DOWN, 
  KEY_LEFT, KEY_RIGHT, KEY_PRINT, KEY_PAUSE,
  //  0, KEY_CAPSLOCK, 0, 0
  0, 0, 0, 0
 };
/* Notice, above, that we are ignoring capslock.  This is because remote
 * commander send shifted keys anyway when capslock is activated.
 * Enabling caplock would have the effect of double-shifting, ie no shifting */


#define NUM_REGULAR_CHARS 128
static unsigned char regular_keycode[NUM_REGULAR_CHARS] = {
  0,0,0,0,0,0,0,0, 
  KEY_BACKSPACE, KEY_TAB, KEY_ENTER, 0,0,0,0,0,
  0,0,0,0,0,0,0,0, 
  0,0,0, KEY_ESC, 0,0,0,0, 
  KEY_SPACE, KEY_1, KEY_APOSTROPHE, KEY_3, 
   KEY_4, KEY_5, KEY_7, KEY_APOSTROPHE, 
  KEY_9, KEY_0, KEY_8, KEY_EQUAL,
   KEY_COMMA, KEY_MINUS, KEY_DOT, KEY_SLASH,
  KEY_0, KEY_1, KEY_2, KEY_3, KEY_4, KEY_5, KEY_6, KEY_7,
  KEY_8, KEY_9, KEY_SEMICOLON, KEY_SEMICOLON,
   KEY_COMMA, KEY_EQUAL, KEY_DOT, KEY_SLASH,
  KEY_2, KEY_A, KEY_B, KEY_C, KEY_D, KEY_E, KEY_F, KEY_G,
  KEY_H, KEY_I, KEY_J, KEY_K, KEY_L, KEY_M, KEY_N, KEY_O, 
  KEY_P, KEY_Q, KEY_R, KEY_S, KEY_T, KEY_U, KEY_V, KEY_W, 
  KEY_X, KEY_Y, KEY_Z, KEY_LEFTBRACE, 
   KEY_BACKSLASH, KEY_RIGHTBRACE, KEY_6, KEY_MINUS,
  KEY_GRAVE, KEY_A, KEY_B, KEY_C, KEY_D, KEY_E, KEY_F, KEY_G,
  KEY_H, KEY_I, KEY_J, KEY_K, KEY_L, KEY_M, KEY_N, KEY_O, 
  KEY_P, KEY_Q, KEY_R, KEY_S, KEY_T, KEY_U, KEY_V, KEY_W, 
  KEY_X, KEY_Y, KEY_Z, KEY_LEFTBRACE, 
   KEY_BACKSLASH, KEY_RIGHTBRACE, KEY_GRAVE, KEY_DELETE
};

static unsigned char keycode_shifts[NUM_REGULAR_CHARS] = {
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 
  0,1,1,1,1,1,1,0, 1,1,1,1,0,0,0,0,
  0,0,0,0,0,0,0,0, 0,0,1,0,1,0,1,1,
  0,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,
  1,1,1,1,1,1,1,1, 1,1,1,0,0,0,1,1,
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0, 0,0,0,1,1,1,1,0
};

struct rcmd {
	unsigned char keycode[128];
	struct input_dev *dev;
	struct serio *serio;
	char phys[32];
};

/* Remote commander messages are very simple:
 * Byte 1: opcode
 * Bytes 2&3: payload size, little endian
 * Byte 4...: payload
 */
enum states { OPCODE, SIZE0, SIZE1, PAYLOAD };
int state = OPCODE;
int payload_size; // we use this variable as a counter when reading payload.
enum opcode { 
  KEY_PRESS, KEY_RELEASE, CTRL_PRESS, CTRL_RELEASE,
  MOUSE_PRESS, MOUSE_RELEASE, MOUSE_MOVE
};
int event_type;

static irqreturn_t rcmd_interrupt(struct serio *serio,
		 unsigned char data, unsigned int flags, struct pt_regs *regs)
{
  struct rcmd *rcmd = serio_get_drvdata(serio);
  
  if( state == OPCODE ){ // new message
    if( data == 0x25 ){
      event_type = KEY_PRESS;
    }else if( data == 0x26 ){
      event_type = KEY_RELEASE;
    }else if( data == 0x27 ){
      event_type = CTRL_PRESS;
    }else if( data == 0x28 ){
      event_type = CTRL_RELEASE;
    }else if( data == 0x23 ){
      event_type = MOUSE_PRESS;
    }else if( data == 0x24 ){
      event_type = MOUSE_RELEASE;
    }else if( data == 0x21 ){
      event_type = MOUSE_MOVE;
    }
    state = SIZE0;     // set next state
  }
  else if( state == SIZE0 ){
    payload_size = data;
    state = SIZE1;     // set next state
  }
  else if( state == SIZE1 ){
    payload_size += data << 8;
    state = PAYLOAD;    // set next state
  }
  else if( state == PAYLOAD ){
    // I have no idea what the following does
    input_regs(rcmd->dev, regs);

    if( event_type == KEY_PRESS ){
      if( data < NUM_REGULAR_CHARS && regular_keycode[data] ){
	// press shift key, if necessary
	if( keycode_shifts[data] ){
	  input_report_key(rcmd->dev, KEY_LEFTSHIFT, 1);
	}
	// press the key
	input_report_key(rcmd->dev, regular_keycode[data], 1); 
      }
    }
    else if( event_type == KEY_RELEASE ){
      if( data < NUM_REGULAR_CHARS && regular_keycode[data] ){
	// release the key
	input_report_key(rcmd->dev, regular_keycode[data], 0);
	// release shift key, if necessary
	if( keycode_shifts[data] ){
	  input_report_key(rcmd->dev, KEY_LEFTSHIFT, 0);
	}	
      }
    }
    else if( event_type == CTRL_PRESS || event_type == CTRL_RELEASE ){
      if( data < NUM_CTRL_CHARS && ctrl_keycode[data] )
	input_report_key(rcmd->dev, ctrl_keycode[data], 
			 event_type == CTRL_PRESS );       
    }
    else if( event_type == MOUSE_PRESS || event_type == MOUSE_RELEASE ){
      if( data == 0 ){ // left click
	input_report_key(rcmd->dev, BTN_LEFT, event_type == MOUSE_PRESS);
      }else if( data == 2 ){ // right click
	input_report_key(rcmd->dev, BTN_RIGHT, event_type == MOUSE_PRESS);
      }
    }
    else if( event_type == MOUSE_MOVE ){
      char offset;
      if( data > 0x80 ){
	offset = data - 0xFF - 1;
      }else{
	offset = data;
      }
      if( payload_size == 2 ){ // first byte is x axis
	input_report_rel(rcmd->dev, REL_X, offset );
      }else if( payload_size == 1 ){ // second byte is y axis
	input_report_rel(rcmd->dev, REL_Y, offset );
      }
    }

    // I don't know what the following does
    input_sync(rcmd->dev);

    payload_size--;
    if( payload_size == 0 ) state = OPCODE; // message is complete, reset state
  }

  return IRQ_HANDLED;  
}


static int rcmd_connect(struct serio *serio, struct serio_driver *drv)
{
	struct rcmd *rcmd;
	struct input_dev *input_dev;
	int err = -ENOMEM;
	int i;

	rcmd = kzalloc(sizeof(struct rcmd), GFP_KERNEL);
	input_dev = input_allocate_device();
	if (!rcmd || !input_dev)
		goto fail;

	rcmd->serio = serio;
	rcmd->dev = input_dev;
	sprintf(rcmd->phys, "%s/input0", serio->phys);
	memcpy(rcmd->keycode, regular_keycode, sizeof(rcmd->keycode));

	input_dev->name = "Remote Commander";
	input_dev->phys = rcmd->phys;
	input_dev->id.bustype = BUS_RS232;
	input_dev->id.vendor = SERIO_REMOTECMD;
	input_dev->id.product = 0x0001;
	input_dev->id.version = 0x0100;
	input_dev->cdev.dev = &serio->dev;
	input_dev->private = rcmd;

	input_dev->evbit[0] = BIT(EV_KEY) | BIT(EV_REP);
	input_dev->keycode = rcmd->keycode;
	input_dev->keycodesize = sizeof(unsigned char);
	input_dev->keycodemax = ARRAY_SIZE(regular_keycode);

	for (i = 0; i < 128; i++){
	  set_bit(i, input_dev->keybit);
	  //	  set_bit(rcmd->keycode[i], input_dev->keybit);
	}
	clear_bit(0, input_dev->keybit);

	// for mouse functionality
	input_dev->evbit[0] |= BIT(EV_KEY) | BIT(EV_REL);
	input_dev->keybit[LONG(BTN_MOUSE)] = BIT(BTN_LEFT) | BIT(BTN_RIGHT);
	input_dev->relbit[0] = BIT(REL_X) | BIT(REL_Y);
	set_bit(BTN_LEFT, input_dev->keybit);
	set_bit(BTN_RIGHT, input_dev->keybit);


	serio_set_drvdata(serio, rcmd);

	err = serio_open(serio, drv);
	if (err)
		goto fail;

	input_register_device(rcmd->dev);
	return 0;

 fail:	serio_set_drvdata(serio, NULL);
	input_free_device(input_dev);
	kfree(rcmd);
	return err;
}

static void rcmd_disconnect(struct serio *serio)
{
	struct rcmd *rcmd = serio_get_drvdata(serio);

	serio_close(serio);
	serio_set_drvdata(serio, NULL);
	input_unregister_device(rcmd->dev);
	kfree(rcmd);
}

static struct serio_device_id rcmd_serio_ids[] = {
	{
		.type	= SERIO_RS232,
		.proto	= SERIO_REMOTECMD,
		.id	= SERIO_ANY,		.extra	= SERIO_ANY,
	},
	{ 0 }
};

MODULE_DEVICE_TABLE(serio, rcmd_serio_ids);

static struct serio_driver rcmd_drv = {
	.driver		= {
		.name	= "remotecmd",
	},
	.description	= DRIVER_DESC,
	.id_table	= rcmd_serio_ids,
	.interrupt	= rcmd_interrupt,
	.connect	= rcmd_connect,
	.disconnect	= rcmd_disconnect,
};

static int __init rcmd_init(void)
{
	serio_register_driver(&rcmd_drv);
	return 0;
}

static void __exit rcmd_exit(void)
{
	serio_unregister_driver(&rcmd_drv);
}

module_init(rcmd_init);
module_exit(rcmd_exit);
